/*
	*
	* Copyright 2019 Britanicus <marcusbritanicus@gmail.com>
	*
	* This file is a part of DesQ project ( https://gitlab.com/desq/ )
	*

	*
	* This program is free software; you can redistribute it and/or modify
	* it under the terms of the GNU General Public License as published by
	* the Free Software Foundation; either version 3 of the License, or
	* (at your option) any later version.
	*

	*
	* This program is distributed in the hope that it will be useful,
	* but WITHOUT ANY WARRANTY; without even the implied warranty of
	* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	* GNU General Public License for more details.
	*

	*
	* You should have received a copy of the GNU General Public License
	* along with this program; if not, write to the Free Software
	* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
	* MA 02110-1301, USA.
	*
*/

#include "DesQSettings.hpp"
#include <QtDBus>

static inline QVariant variantMapToVariant( QVariantMap valueMap ) {

	QVariant value = valueMap[ "value" ];
	int type = (int)valueMap[ "type" ].toInt();

	QDBusArgument dbusArg = value.value<QDBusArgument>();
	switch( type ) {
		case QMetaType::QStringList: {
			QStringList list;
			dbusArg >> list;
			return list;
		}

		case QMetaType::QRect: {
			QRect rect;
			dbusArg >> rect;
			return rect;
		}

		case QMetaType::QSize: {
			QSize size;
			dbusArg >> size;
			return size;
		}

		default: {
			qDebug() << dbusArg.currentType();
			return QVariant();
		}
	}

	return QVariant();
};

DesQSettings::DesQSettings( QString app ) {

	QProcess::startDetached( "/usr/libexec/desq/desq-settings-daemon", { "-d" } );

	mAppName = app;
	fallback = new QSettings( ":/Settings/DesQ" + app + ".conf", QSettings::IniFormat );

	iface = new QDBusInterface( "org.desq.Settings", "/org/DesQ/Settings", "org.desq.Settings", QDBusConnection::sessionBus() );

	QDBusConnection::sessionBus().connect(
		"org.desq.Settings",
		"/org/DesQ/Settings",
		"org.desq.Settings",
		"settingChanged",
		"a{sv}",
		this,
		SLOT( emitSettings( QVariantMap ) )
	);

	if ( not iface->isValid() ) {
		qDebug() << "Unable to get desq-settings-daemon running.";
		qDebug() << "We will use a static fallback. You experience may be very limited.";
		qDebug() << "If you feel this is a bug, open an issue at https://gitlab.com/DesQ/Settings/issues/new.";
	}
};

void DesQSettings::reload() {

	if ( not iface->isValid() ) {
		qDebug() << "Using a static fallback. You experience may be very limited.";
		fallback->sync();
		return;
	}

	iface->call( "reload" );
};

QVariant DesQSettings::rawValue( const QString &key ) {

	if ( not fallback->contains( key ) )
		return QVariant();

	if ( not iface->isValid() ) {
		qDebug() << "Using a static fallback. You experience may be very limited.";

		QVariant value;

		/* Here we can define some specialized/dynamic settings: Example form factor */
		if ( key == "FormFactor" ) {

			value = QVariant( 1 );
		}

		/* TouchMode: As seen by the oser */
		else if ( key == "TouchMode" ) {
			int touch = fallback->value( key ).toInt();
			switch ( touch ) {
				/* Auto detect */
				case 0: {
					QProcess proc;
					proc.start( "udevadm", QStringList() << "info" << "--export-db" );
					proc.waitForFinished( -1 );

					QString output = QString::fromLocal8Bit( proc.readAllStandardOutput() + '\n' + proc.readAllStandardError() );
					if ( output.contains( "ID_INPUT_TOUCHSCREEN=1" ) )
						value = QVariant( true );

					value = QVariant( false );
					break;
				}

				/* TouchMode is on */
				case 1:
					value = QVariant( true );
					break;

				/* TouchMode is off */
				case 2:
					value = QVariant( false );
					break;

				default:
					value = QVariant( false );
					break;
			}
		}

		/* TouchModeRaw: Raw value saved in the file */
		else if ( key == "TouchModeRaw" ) {

			value = fallback->value( "TouchMode" );
		}

		/* Get the default value for the key */
		else {
			value = fallback->value( key );

			if ( (QMetaType::Type)value.type() == QMetaType::QString ) {
				QString strValue = value.toString();
				if ( strValue.startsWith( "~" ) )
					strValue = strValue.replace( "~", QDir::homePath() );

				value = QVariant( strValue );
			}
		}

		return value;
	}

	QDBusReply<QVariantMap> dbusVar = iface->call( "value", mAppName, key );
	QVariant value = dbusVar.value()[ "value" ];

	if ( value.type() >= 1024 )
		value = variantMapToVariant( dbusVar.value() );

	if ( (QMetaType::Type)value.type() == QMetaType::QString ) {
		QString str = value.toString();
		if ( str.startsWith( "QFont\n" ) ) {
			QStringList font = str.split( "\n" );
			QString name = font.at( 1 );
			int pointSize = font.at( 2 ).toInt();
			int weight = font.at( 3 ).toInt();
			bool italic = ( font.at( 4 ) == "true" ? true : false );

			value = QVariant( QFont( name, pointSize, weight, italic ) );
		}
	}

	return value;
};

DesQSettings::DesQProxy DesQSettings::value( const QString &key ) {

	QVariant value = rawValue( key );
	return DesQProxy{ value };
};

void DesQSettings::setValue( const QString &key, QVariant value ) {

	if ( not iface->isValid() ) {
		qDebug() << "Using a static fallback. These settings may not be saved properly";
		fallback->setValue( key, value );

		return;
	}

	if ( ( QMetaType::Type )value.type() == QMetaType::QFont ) {
		QFont f = value.value<QFont>();
		value = QString( "QFont\n%1\n%2\n%3\n%4" ).arg( f.family() ).arg( f.pointSize() ).arg( f.weight() ).arg( f.italic() );
	}

	QVariantMap map;

	map[ "app" ] = mAppName;
	map[ "key" ] = key;
	map[ "value" ] = value;
	map[ "type" ] = (int)value.type();

	iface->call( "setValue", map );
};

void DesQSettings::emitSettings( QVariantMap valueMap ) {

	QString app = valueMap[ "app" ].toString();
	QString key = valueMap[ "key" ].toString();
	QVariant value = valueMap[ "value" ];

	if ( value.type() >= 1024 )
		value = variantMapToVariant( valueMap );

	if ( (QMetaType::Type)value.type() == QMetaType::QString ) {
		QString str = value.toString();
		if ( str.startsWith( "QFont\n" ) ) {
			QStringList font = str.split( "\n" );
			QString name = font.at( 1 );
			int pointSize = font.at( 2 ).toInt();
			int weight = font.at( 3 ).toInt();
			bool italic = ( font.at( 4 ) == "true" ? true : false );

			value = QVariant( QFont( name, pointSize, weight, italic ) );
		}
	}

	emit settingChanged( app, key, value );
};
